# Copyright (c) 2017  University of Houston

from __future__ import absolute_import

import socket

import IPy

from foam.geni.resources import SkipHandler
from foam.vts.images import ImageHandler
from foam.vts.drivers.docker import DockerDriver

def _wrapPOA (name, klass, methname):
  def _poaRedirect (action, urns, options, user_urn):
    from foam.vts.data import DB, UnknownInstanceError
    from foam.vts.geni import SliverBundleProxy
    from foam.geni.resources import UnknownSliceError, UnsupportedResourceActionError

    if len(urns) > 1:
      return {}

    try:
      sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
    except KeyError:
      raise UnknownSliceError(urns[0])

    retl = {}
    for clid in options["client-ids"]:
      ct = sliver.getContainer(clid)
      if not ct:
        continue
      if not ct._instance_handler:
        raise UnsupportedResourceActionError(name, clid)
      if isinstance(ct._instance_handler, klass):
        meth = getattr(ct._instance_handler, methname)
        data = meth(options, user_urn)
        retl[clid] = data
    if not retl:
      # We registered this handler, but none of our client-ids were passed in
      raise SkipHandler()
    return retl

  _poaRedirect._realinfo = ("%s.%s" % (klass.__module__, klass.__name__), methname)
  return (name, _poaRedirect)

def _wrapMapPOA (name, klass, methname):
  def _poaRedirect (action, urns, options, user_urn):
    from foam.vts.data import DB, UnknownInstanceError
    from foam.vts.geni import SliverBundleProxy
    from foam.geni.resources import UnknownSliceError, UnsupportedResourceActionError

    if len(urns) > 1:
      return {}

    try:
      sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
    except KeyError:
      raise UnknownSliceError(urns[0])

    retl = {}
    for clid,data in options["client-id-map"].items():
      ct = sliver.getContainer(clid)
      if not ct:
        raise UnknownInstanceError(clid, "")
      if not ct._instance_handler:
        raise UnsupportedResourceActionError(name, clid)
      if isinstance(ct._instance_handler, klass):
        meth = getattr(ct._instance_handler, methname)
        ret = meth(options, data, user_urn)
        retl[clid] = ret
    if not retl:
      # We registered this handler, but none of our client-ids were passed in
      raise SkipHandler()
    return retl

  _poaRedirect._realinfo = ("%s.%s" % (klass.__module__, klass.__name__), methname)
  return (name, _poaRedirect)

def arphwtype (data):
  # We only bother to decode a few here that we're likely to see
  # The full list is defined in linux/if_arp.h
  val = int(data, 16)
  if val == 1:
    return "ethernet"
  elif val == 24:
    return "firewire"
  elif val == 32:
    return "infiniband"
  elif val == 801:
    return "802.11"
  else:
    return "unknown (%d)" % (val)

def arpstatus (data):
  # This is not really the meaning of the flags field, but we try
  # to simplify it.  I'm sure this will never come back and bite us.
  val = int(data, 16)
  if val == 0:
    return "incomplete"
  elif val == 2:
    return "complete"
  elif val == 6:
    return "static"
  else:
    return "unknown (%d)" % (val)

def ipstr (data):
  val = socket.ntohl(int(data, 16))
  return str(IPy.IP(val))


class BadCommandError(Exception):
  def __init__ (self, cmd):
    self.cmd = cmd
  def __str__ (self):
    return "Command string [%s] rejected by runtime sanitizer" % (self.cmd)

BAD_FIRST_CHARS = set(["-", "\\", ";"])
def sanitize (cmd):
  # This isn't so much intended to be secure as to prevent people from making
  # particularly egregious mistakes or typos / missing format strings that
  # screw us
  cmd = cmd.strip()
  if cmd[0] in BAD_FIRST_CHARS:
    raise BadCommandError(cmd)
  return cmd


class AlpineHost(ImageHandler):
  def _logOutput (self, o, e, r):
    if o.strip(): self.log.info(o)
    if e.strip(): self.log.debug(e)
    return (o, e, r)

  def addUser (self, options, user_urn):
    outl = []
    dd = DockerDriver(self.log.debug)

    (o,e,r) = dd.ccall(self.container, "adduser -s %s -D %s" % (options["shell"], options["username"]))
    outl.append({"adduser" : (o, e, r)})

    if options.has_key("password"):
      (o,e,r) = self._logOutput(*dd.ccall(self.container, "sh -c \"echo '%s:%s' | chpasswd\"" % (options["username"], options["password"])))
      outl.append({"chpasswd" : (o, e, r)})

    if options.has_key("ssh-pubkey"):
      self._logOutput(*dd.ccall(self.container, "sh -c \"mkdir ~%s/.ssh/\"" % (options["username"])))
      self._logOutput(*dd.ccall(self.container, "sh -c \"echo '%s' > ~%s/.ssh/authorized_keys\"" % (options["ssh-pubkey"], options["username"])))
      self._logOutput(*dd.ccall(self.container, "sh -c \"chown -R %s:%s ~%s/.ssh\"" % (options["username"], options["username"], options["username"])))

    return outl

  def getARPTable (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    rawtable = dd.getProcData(self.container, "net/arp")
    ret = []
    for line in rawtable[1:]:
      o = line.split()
      obj = {}
      obj["ip-address"] = o[0]
      obj["hw-type"] = arphwtype(o[1])
      obj["status"] = arpstatus(o[2])
      obj["hw-address"] = o[3]
      obj["device"] = o[5]
      ret.append(obj)
    return ret

  def getRouteTable (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    rawtable = dd.getProcData(self.container, "net/route")
    ret = []
    for line in rawtable[1:]:
      o = line.split()
      obj = {}
      obj["device"] = o[0]
      obj["destination"] = ipstr(o[1])
      obj["gateway"] = ipstr(o[2])
      obj["mask"] = ipstr(o[7])
      ret.append(obj)
    return ret

  def execCmd (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    cmd = sanitize(options["cmd"])
    (o,e,r) = dd.ccall(self.container, cmd)
    return {"cmd" : cmd, "stdout" : o, "stderr" : e, "retcode" : r}

  @staticmethod
  def geniSetup (registry):
    registry.addPOACallback(*(AlpineHost.poaGetARPTable))
    registry.addPOACallback(*(AlpineHost.poaGetRouteTable))
    registry.addPOACallback(*(AlpineHost.poaAddUser))
#    registry.addPOACallback(*(AlpineHost.poaExec))

AlpineHost.poaGetARPTable = _wrapPOA("api:uh.host:get-arp-table", AlpineHost, "getARPTable")
AlpineHost.poaGetRouteTable = _wrapPOA("api:uh.host:get-route-table", AlpineHost, "getRouteTable")
AlpineHost.poaAddUser = _wrapPOA("api:uh.host:add-user", AlpineHost, "addUser")
#AlpineHost.poaExec = _wrapPOA("api:uh.host:exec", AlpineHost, "execCmd")


class SupervisorHost(ImageHandler):
  def supStatus (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.supervisor(self.container, "status", None)
    return o

  def supAvail (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.supervisor(self.container, "avail", None)
    return o

  def supStop (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.supervisor(self.container, "stop", options["svc-name"])
    return o

  def supStart (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.supervisor(self.container, "start", options["svc-name"])
    return o

  def supRestart (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.supervisor(self.container, "restart", options["svc-name"])
    return o

  def supUpdate (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.supervisor(self.container, "update", None)
    return o

  @staticmethod
  def geniSetup (registry):
    registry.addPOACallback(*(SupervisorHost.poaSupStatus))
    registry.addPOACallback(*(SupervisorHost.poaSupAvail))
    registry.addPOACallback(*(SupervisorHost.poaSupStop))
    registry.addPOACallback(*(SupervisorHost.poaSupStart))
    registry.addPOACallback(*(SupervisorHost.poaSupRestart))
    registry.addPOACallback(*(SupervisorHost.poaSupUpdate))


SupervisorHost.poaSupStatus = _wrapPOA("api:uh.host:supervisor-status", SupervisorHost, "supStatus")
SupervisorHost.poaSupAvail = _wrapPOA("api:uh.host:supervisor-avail", SupervisorHost, "supAvail")
SupervisorHost.poaSupStop = _wrapPOA("api:uh.host:supervisor-stop", SupervisorHost, "supStop")
SupervisorHost.poaSupStart = _wrapPOA("api:uh.host:supervisor-start", SupervisorHost, "supStart")
SupervisorHost.poaSupRestart = _wrapPOA("api:uh.host:supervisor-restart", SupervisorHost, "supRestart")
SupervisorHost.poaSupUpdate = _wrapPOA("api:uh.host:supervisor-update", SupervisorHost, "supUpdate")

class AlpineSupervisorHost(AlpineHost, SupervisorHost):
  @staticmethod
  def geniSetup (registry):
    AlpineHost.geniSetup(registry)
    SupervisorHost.geniSetup(registry)
