# Copyright (c) 2016-2017  University of Houston

from __future__ import absolute_import

import datetime
import os
import shutil

from .bases import _wrapPOA, _wrapMapPOA

from foam.core.exception import CoreException
from foam.vts.images import ImageHandler
from foam.vts.hostdriver import HostDriver
from foam.vts.data import DB, UnknownInstanceError
from foam.vts.drivers.docker import DockerDriver, HostPathVolume
from foam.geni.resources import UnknownSliceError, UnsupportedResourceActionError
from IPy import IP

STORE = "/tmp/dns-dhcp/%s/%s"

DNSMASQCONF = """server=%(rootDNSIP)s
local=/%(subDomainName)s/
expand-hosts
domain=%(subDomainName)s
cache-size=150
no-negcache
no-hosts
addn-hosts=/data/add_hosts
dhcp-range=%(start-ip)s, %(end-ip)s, %(lease-time)d
dhcp-option=option:router,%(routerIP)s
dhcp-leasefile=/data/dnsmasq.leases
"""

ADDHOSTSFILE = """%(ipAddress1)s    %(vmName)s.%(subDomainName)s  %(vmName)s
%(ipAddress2)s     %(vmName)s.%(subDomainName)s  %(vmName)s
127.0.0.1       localhost
::1     localhost ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
"""

LOGROTATECONF = """/data/dnsmasq.log {
    size 1M
    create 0664 root root
    rotate 1
    maxage 2
    notifempty
}
"""


class DnsDhcpConfigError(CoreException):
  def __init__(self, client_id, errorDescription):
    super(DnsDhcpConfigError, self).__init__()
    self.cid = client_id
    self.description = errorDescription

  def __str__(self):
    return "error in configuring uh.dnsdhcp node [(%s)]: (%s) " % (self.cid, self.description)


class DNSDHCPHandler(ImageHandler):
  def prebuild(self, cobj):
    '''touch log & lease files, mount in /data'''
    self.log.debug("Prebuild (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
    os.makedirs(dpath)
    hd = HostDriver(self.log.info)
    hd.touch("%s/dnsmasq.log" % (dpath))
    hd.touch("%s/dnsmasq.leases" % (dpath))
    self.volumes.append(HostPathVolume(dpath, "/data"))

  def _writeConfigs(self, dpath, cobj):
    '''write dnsmasq.conf, add_hosts files, and logrotate.conf'''
    self.log.debug("_writeConfigs (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    configpath = "%s/dnsmasq.conf" % (dpath)
    addhostspath = "%s/add_hosts" % (dpath)
    logrotateconfigpath = "%s/logrotate.conf" % (dpath)
    configx = DNSMASQCONF
    addhostsfile = ADDHOSTSFILE
    rootDNSIP = ""
    dotRootDomain = ""
    ipPortToRoot = ""
    vmName = cobj.hostname
    subnet = self.attrs["subnet"]
    ipDhcpPort = str(subnet[1]).split('/')[0]
    if "root-dns" not in self.attrs or self.attrs["root-dns"] == None:
      # congig if there is no rootdns specified.
      configx = DNSMASQCONF.split('\n', 1)[1]
      addhostsfile = ADDHOSTSFILE.split('\n', 4)[4]
    else:
      # else config if there is a rootdns specified.\
      if "interface-to-root" not in self.attrs or self.attrs["interface-to-root"] == None:
        raise DnsDhcpConfigError(cobj.client_id, 'attribute [interface-to-root] needs to be set')
      if "interface-for-dhcp" not in self.attrs or self.attrs["interface-for-dhcp"] == None:
        raise DnsDhcpConfigError(cobj.client_id, 'attribute [interface-for-dhcp] needs to be set')
      if (self.attrs["interface-to-root"] != self.attrs["interface-for-dhcp"]):
        # for out of band root dns server
        port_to_root = cobj.getPort(self.attrs["interface-to-root"])
        try:
          ipPortToRoot = port_to_root.ipv4addrs[0].split('/')[0]  # returns a list: take the first.
        except Exception:
          raise DnsDhcpConfigError(cobj.client_id, 'IPv4 Address of [port-to-root] not set')
      else:
        # for in band rootdns server
        ipPortToRoot = ipDhcpPort
      port_of_root = None
      sliver = cobj._sliver
      for container in sliver.containers:
        if container.imagename == "uh.dnsroot":
          if self.attrs["root-dns"] == container.client_id:
            # port_of_root = container._instance_handler.attrs["port-of-root-dns"]
            port_of_root = container.ports.values()[0]
            if container._instance_handler.attrs["top-domain-name"]:
              rootDomainName = container._instance_handler.attrs["top-domain-name"].strip(' .')
              dotRootDomain = "." + rootDomainName
      try:
        rootDNSIP = port_of_root.ipv4addrs[0].split('/')[0]  # returns a list: take the first.
      except Exception:
        raise DnsDhcpConfigError(cobj.client_id, 'IPv4 Address of [port_of_root] was not set')
    with open(configpath, "w+") as f:
      f.write(configx % {"rootDNSIP": rootDNSIP,
                         "subDomainName": self.attrs["sub-domain-name"].strip(" .") + dotRootDomain,
                         "start-ip": str(subnet[10]),
                         "end-ip": str(subnet[-6]),
                         "lease-time": self.attrs["lease-duration"],
                         "routerIP": str(subnet[-2])})
    with open(addhostspath, "w+") as f:
      f.write(addhostsfile % {"ipAddress1": ipDhcpPort,
                              "ipAddress2": ipPortToRoot,
                              "subDomainName": self.attrs["sub-domain-name"].strip(" .") + dotRootDomain,
                              "vmName": str(vmName)})
    with open(logrotateconfigpath, "w+") as f:
      f.write(LOGROTATECONF)

  def postlocaltopo(self, cobj):
    '''set ip addresses of dhcp interface'''
    self.log.debug("postlocaltopo (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    subnet = self.attrs["subnet"]
    dhcpPort = cobj.getPort(self.attrs["interface-for-dhcp"])
    self.setIPAddress(cobj, dhcpPort.alias, str(subnet[1]), subnet.prefixlen())

  def postnettopo(self, cobj):
    '''write config files and restart dnsmasq'''
    self.log.debug("postnettopo (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
    self._writeConfigs(dpath, cobj)
    dd = DockerDriver(self.log.info)
    dd.supervisor(cobj, "start", "dnsmasq")

  def postrm(self, cobj):
    '''shut down the image'''
    self.log.debug("postrm (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    try:
      dpath = STORE % (cobj._sliver.uuid, "")
      shutil.rmtree(dpath)
    except OSError:
      self.log.info("Could not delete %s" % (dpath))

  def _reWriteConfigs(self, dpath, vmName):
    '''re-write dnsmasq.conf file using new subnet details'''
    self.log.debug("_reWriteConfigs (%s)" % (vmName))
    configpath = "%s/dnsmasq.conf" % (dpath)
    addhostspath = "%s/add_hosts" % (dpath)
    subnet = self.attrs["subnet"]
    confstring = ""
    with open(configpath, 'r') as file:
      for row in file:
        if "dhcp-range" in row:
          leasetime = row.split(',')[2].strip('\n')
          confstring += "dhcp-range = " + str(subnet[10]) + "," + str(subnet[-6]) + "," + leasetime + "\n"
        elif "router" in row:
          routerconf = row.split(',')[0]
          confstring += routerconf + "," + str(subnet[-2]) + "\n"
        else:
          confstring += row
    with open(configpath, "w+") as f:
      f.write(confstring)
    addhoststring = ""
    found = False
    with open(addhostspath, 'r') as file:
      for row in file:
        if (vmName in row) and not found:
          names = row.split(' ', 1)[1]
          addhoststring += str(subnet[1]) + " " + names
          found = True
        else:
          addhoststring += row
    with open(addhostspath, "w+") as f:
      f.write(addhoststring)

  def getLeases(self, options, user_urn):
    try:
      self.log.debug("getLeases (%s, %s)" % (self.container._sliver.uuid, self.container.client_id))
      leasesfile = "%s/dnsmasq.leases" % (STORE % (self.container._sliver.uuid, self.container.client_id))
      retl = []
      with open(leasesfile, 'r') as file:
        for row in file:
          expr_time, mac_addr, ip_addr, hostname, ex = row.split()
          expr_time = datetime.datetime.fromtimestamp(float(expr_time))
          # Start cannot be a "*", but we'll let that screw us later
          retl.append({"hostname": hostname, "mac-address": mac_addr,
                       "ip-address": ip_addr, "binding-state": "*",
                       "start": "*", "end": expr_time, "options": ""})
      return retl
    except Exception as e:
      self.log.exception("unhandled exception while getting leases \n" + str(e))
      return []

  def setSubnet(self, options, data, user_urn):
    '''set subnet using _reWriteConfigs()'''
    subnet = IP(data)
    try:
      self.log.debug("setSubnet (%s, %s)" % (self.container._sliver.uuid, self.container.client_id))
      self.attrs["subnet"] = subnet
      dpath = STORE % (self.container._sliver.uuid, self.container.client_id)
      self._reWriteConfigs(dpath, self.container.hostname)
      dhcpPort = self.container.getPort(self.attrs["interface-for-dhcp"])
      dhcpInterface = dhcpPort.alias
      self.setIPAddress(self.container, dhcpInterface, str(subnet[1]), subnet.prefixlen())
      dd = DockerDriver(self.log.info)
      dd.supervisor(cobj, "restart", "dnsmasq")
      return True
    except Exception as e:
      self.log.exception("unhandled exception while setting subnet \n" + str(e))
      return False

  def getLastDNSDHCPops(self, cobj, numberOfOps, dnsORdhcp):
    '''pulls last 20 dns related entries logs fromdnsmasq.logs '''
    try:
      self.log.debug("getLastDNSops (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
      logfile = "%s/dnsmasq.log" % (STORE % (cobj._sliver.uuid, cobj.client_id))
      retl = []
      count = 0
      for line in reversed(open(logfile).readlines()):
        if dnsORdhcp == "dns":
          if (" dnsmasq[" in line):
            retl.insert(0, {"data" : line})
            count += 1
        elif dnsORdhcp == "dhcp":
          if ("dnsmasq-dhcp[" in line):
            retl.insert(0, {"data" : line})
            count += 1
        else:
          retl.insert(0, {"data" : line})
          count += 1
        if count == numberOfOps:
          break
      return retl
    except Exception:
      self.log.exception("Unhandled exception in getLastDNSDHCPops")
      return []

  @staticmethod
  def geniSetup(registry):
    registry.addPOACallback(*(DNSDHCPHandler.poaGetLeases))
    registry.addPOACallback(*(DNSDHCPHandler.poaSetSubnet))
    registry.addPOACallback("vts:uh.dnsdhcp:get-last-DNSDHCP-ops", poaGetLastDNSDHCPops)

DNSDHCPHandler.poaGetLeases = _wrapPOA("api:uh.dhcp:get-leases", DNSDHCPHandler, "getLeases")
DNSDHCPHandler.poaSetSubnet = _wrapMapPOA("api:uh.dhcp:set-subnet", DNSDHCPHandler, "setSubnet")


def poaGetLastDNSDHCPops(action, urns, options, user_urn):
  from foam.vts.geni import SliverBundleProxy

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  retl = {}
  for clid in options["client-ids"]:
    ct = sliver.getContainer(clid)
    if not ct:
      raise UnknownInstanceError(clid, "uh.dnsdhcp")
    if isinstance(ct._instance_handler, DNSDHCPHandler):
      lastqueries_data = ct._instance_handler.getLastDNSDHCPops(ct, options["number-of-operations"], options["dns-OR-dhcp"])
      retl[clid] = lastqueries_data
    else:
      raise UnsupportedResourceActionError("vts:uh.dnsdhcp:get-last-DNSDHCP-ops", clid)
  return retl

  # TODO ask if we should automatically update the dnsroot with new IP on set subnet
