# Copyright (c) 2017  University of Houston

from __future__ import absolute_import

from foam.vts.drivers.docker import DockerDriver

from .bases import AlpineSupervisorHost, _wrapPOA
from .util import RSYSLOGCONF

"""
Codes: K - kernel route, C - connected, S - static, R - RIP,
       O - OSPF, I - IS-IS, B - BGP, P - PIM, A - Babel,
       > - selected route, * - FIB route

O   10.0.50.0/24 [110/10] is directly connected, eth1, 00:00:43
C>* 10.0.50.0/24 is directly connected, eth1
O   10.0.70.0/24 [110/10] is directly connected, eth2, 00:00:11
C>* 10.0.70.0/24 is directly connected, eth2
C>* 127.0.0.0/8 is directly connected, lo

K>* 10.0.0.0/24 via 192.168.0.1, eth1
C>* 10.0.1.0/24 is directly connected, eth2
K>* 10.50.0.0/24 via 192.168.0.3, eth1
C>* 127.0.0.0/8 is directly connected, lo
C>* 192.168.0.0/25 is directly connected, eth1
"""

class Quagga(AlpineSupervisorHost):
  def postlocaltopo (self, cobj):
    hvol = None
    for vol in self.container._vols:
      if vol.cpath == "/cfg":
        hvol = vol
        break

    open("%s/ospfd.conf" % (hvol.hpath), "a").close()
    open("%s/ripd.conf" % (hvol.hpath), "a").close()
    open("%s/bgpd.conf" % (hvol.hpath), "a").close()
    open("%s/zebra.conf" % (hvol.hpath), "a").close()

    if not self.attrs["syslog-ip"]:
      return

    data = {"syslog-ip" : self.attrs["syslog-ip"], "work-dir" : "/cfg"}
    with open("%s/rsyslog.conf" % (hvol.hpath), "w+") as cfgf:
      cfgf.write(RSYSLOGCONF % (data))

  def postnettopo (self, cobj):
    dd = DockerDriver(self.log.info)
    dd.ccall(self.container, "chown -R quagga:quagga /cfg")

    if self.attrs["syslog-ip"]:
      dd.supervisor(self.container, "start", "rsyslog")

    dd.supervisor(self.container, "start", "zebra")
    dd.supervisor(self.container, "start", "bgp")
    dd.supervisor(self.container, "start", "rip")
    dd.supervisor(self.container, "start", "ospf")

    if self.attrs["syslog-ip"]:
      dd.ccall(self.container, 'vtysh -c "conf t" -c "log syslog debugging"')
      dd.ccall(self.container, 'vtysh -c "write mem"')
    

  def getIPRouteTable (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.ccall(self.container, 'vtysh -c "show ip route"')
    objlist = Quagga.parseRouteTable(o)
    return objlist

  def getOSPFNeighbors (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    (o,e,r) = dd.ccall(self.container, 'vtysh -c "show ip ospf neighbor"')

    retlist = []
    lines = o.split("\n")
    for line in lines[2:]:
      if not line.strip():
        continue
      nbr = {}
      parts = line.split()
      nbr["id"] = parts[0]
      nbr["priority"] = parts[1]
      nbr["state"] = parts[2]
      nbr["dead-time"] = parts[3]
      nbr["address"] = parts[4]
      nbr["interface"] = parts[5]
      retlist.append(nbr)
    return retlist

  def addOSPFNetworks (self, options, user_urn):
    dd = DockerDriver(self.log.debug)
    # options:
    # { "client_ids" : [<client-id>, ...],
    #   "networks" : [(<network>, <area>), ... ] }
    nets = options["networks"]
    for (network,area) in nets:
      dd.ccall(self.container, 'vtysh -c "conf t" -c "router ospf" -c "network %s area %s"' % (network, area))
    return None

  @staticmethod
  def parseRouteTable (text):
    data = []
    lines = text.split("\n")

    # TODO:  This is totally inadequate for anything that isn't super trivial

    # We start at line 5, to skip all the header codes
    for line in lines[4:]:
      if not line.strip():
        continue
      if line[0] == "C":
        data.append(Quagga.parseConnected(line))
        last_proto = "C"
      elif line[0] == "K":
        data.append(Quagga.parseKernel(line))
        last_proto = "K"
      elif line[0] == "O":
        data.append(Quagga.parseOSPF(line))
        last_proto = "O"
      elif line[0] == " ":
        if last_proto == "O":
          data.append(Quagga.parseOSPFContinuation(line))
      else:
        last_proto = None
    return data

  @staticmethod
  def parseConnected (line):
    obj = {"type" : "C", "selected" : False, "fib" : False, "time" : None}
    if line[1] == ">":
      obj["selected"] = True
    if line[2] == "*":
      obj["fib"] = True
    parts = line[3:].split(",")

    rte_parts = parts[0].split()
    obj["network"] = rte_parts[0]
    obj["next-hop"] = " ".join(rte_parts[2:]) # Strip the "is"

    obj["interface"] = parts[1].strip()
    return obj

  @staticmethod
  def parseKernel (line):
    obj = {"type" : "K", "selected" : False, "fib" : False, "time" : None}
    if line[1] == ">":
      obj["selected"] = True
    if line[2] == "*":
      obj["fib"] = True

    parts = line[3:].split(",")
    rte_parts = parts[0].split()
    obj["network"] = rte_parts[0]
    obj["next-hop"] = rte_parts[2] # 1 is "via"

    obj["interface"] = parts[1].strip()
    return obj

  @staticmethod
  def parseOSPF (line):
    obj = {"type" : "O", "selected" : False, "fib" : False, "time" : None}
    if line[1] == ">":
      obj["selected"] = True
    if line[2] == "*":
      obj["fib"] = True
    parts = line[3:].split(",")

    rte_parts = parts[0].split()
    obj["network"] = " ".join(rte_parts[0:2])
    obj["next-hop"] = " ".join(rte_parts[3:])

    obj["interface"] = parts[1].strip()
    obj["time"] = parts[2].strip()
    return obj

  @staticmethod
  def parseOSPFContinuation (line):
    #  *                         via 172.16.14.2, eth4, 00:00:17
    obj = {"type" : "", "selected" : False, "fib" : False, "time" : None}
    if line[2] == "*":
      obj["fib"] = True

    parts = line[3:].split(",")
    obj["network"] = ""
    obj["next-hop"] = parts[0].strip().split()[1]
    obj["interface"] = parts[1].strip()
    obj["time"] = parts[2].strip()
    return obj

  @staticmethod
  def geniSetup (registry):
    AlpineSupervisorHost.geniSetup(registry)

    registry.addPOACallback(*(Quagga.poaRouteTable))
    registry.addPOACallback(*(Quagga.poaAddOSPFNets))
    registry.addPOACallback(*(Quagga.poaOSPFNeighbors))

Quagga.poaRouteTable = _wrapPOA("vts:uh.quagga:get-route-table", Quagga, "getIPRouteTable")
Quagga.poaAddOSPFNets = _wrapPOA("vts:uh.quagga:add-ospf-nets", Quagga, "addOSPFNetworks")
Quagga.poaOSPFNeighbors = _wrapPOA("vts:uh.quagga:get-ospf-neighbors", Quagga, "getOSPFNeighbors")

