# Copyright (c) 2016  Barnstormer Softworks, Ltd.
# Copyright (c) 2016-2018  University of Houston

from __future__ import absolute_import

import datetime
import json
import os
import stat
import shutil

from isc_dhcp_leases.iscdhcpleases import IscDhcpLeases
import IPy

from .bases import AlpineHost, SupervisorHost, _wrapPOA, _wrapMapPOA
from .util import RSYSLOGCONF

from foam.vts.drivers.docker import DockerDriver, HostPathVolume
from foam.vts.hostdriver import HostDriver
from foam.vts.data import UnknownInstanceError

STORE = "%s/simple-dhcp/%s/%s"
HDATA = "%s/simple-dhcp/%s/%s.hdata"

BASE_CONFIG = """
default-lease-time %(lease-time)d;
authoritative;
"""

SUBNET_CONFIG = """
subnet %(network)s netmask %(netmask)s {
  range %(start-ip)s %(end-ip)s;
  %(router-option)s
  %(dns-option)s
}"""

HOST_CONFIG = """
host %(name)s {
  hardware ethernet %(mac)s;
  fixed-address %(addr)s;
}"""

CONFIG = "%(base)s\n%(subnets)s\n%(hosts)s"

class SimpleDHCPHandler(AlpineHost, SupervisorHost):
  ATTRS = { "intf.dhcpv4" : False,
            "intf.dhcpv6" : False,
            "intf.staticv4" : True,
            "intf.staticv6" : False
          }

  def prebuild (self, cobj):
    self.log.debug("Prebuild (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))

    hd = HostDriver(self.log.info)

    dpath = STORE % (hd.volroot, cobj._sliver.uuid, cobj.client_id)
    hpath = HDATA % (hd.volroot, cobj._sliver.uuid, cobj.client_id)

    os.makedirs(dpath)
    cpath = "%s/dhcpd.conf" % (dpath)
    rpath = "%s/dhcpd" % (dpath)

    self._writeConfig(hpath, cpath)

    hd.touch("%s/dhcpd.leases" % (dpath))

    self.volumes.append(HostPathVolume(dpath, "/data"))

  def _writeHosts (self, hpath, hinfo):
    with open(hpath, "w+") as hfile:
      hfile.write(json.dumps(hinfo))

  def _writeConfig (self, hpath, cpath, hinfo = None):
    if hinfo is None:
      hinfo = {}

    if (not os.path.exists(hpath)) or hinfo:
      self._writeHosts(hpath, hinfo)
    else:
      hinfo = json.loads(open(hpath, "r").read())

    hstrs = [HOST_CONFIG % (v) for k,v in hinfo.items()]
    subnet = self.attrs["subnet"]
    passone = {"netmask" : str(subnet.netmask()),
               "network" : str(subnet.net()),
               "start-ip" : str(subnet[self.attrs["reserved-ips"]]),
               "end-ip" : str(subnet[-3]),
               "lease-time" : self.attrs["lease-duration"],
               "router-option" : "",
               "dns-option" : ""}
    if self.attrs["has-router"]:
      passone["router-option"] = "option routers %s; " % (str(subnet[-2]))
    if self.attrs["dns-ip"]:
      passone["dns-option"] = "option domain-name-servers %s;" % (str(self.attrs["dns-ip"]))

    basestr = BASE_CONFIG % (passone)
    subnetstr = SUBNET_CONFIG % (passone)

    passtwo = {"base" : basestr, "subnets" : subnetstr, "hosts" : "\n".join(hstrs)}

    with open(cpath, "w+") as f:
      f.write(CONFIG % (passtwo))

  def postlocaltopo (self, cobj):
    from foam.vts.drivers.docker import DockerDriver
    subnet = self.attrs["subnet"]
    self.setIPAddress(cobj, "eth1", subnet[1], subnet.prefixlen())

    if not self.attrs["syslog-ip"]:
      return

    hvol = None
    for vol in self.container._vols:
      if vol.cpath == "/cfg":
        hvol = vol
        break

    data = {"syslog-ip" : self.attrs["syslog-ip"], "work-dir" : "/cfg"}
    with open("%s/rsyslog.conf" % (hvol.hpath), "w+") as cfgf:
      cfgf.write(RSYSLOGCONF % (data))

  def postnettopo (self, cobj):
    dd = DockerDriver(self.log.info)

    if self.attrs["syslog-ip"]:
      dd.supervisor(self.container, "start", "rsyslog")

    dd.supervisor(self.container, "start", "dhcpd")


  def postrm (self, cobj):
    self.log.debug("postrm (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))

    hd = HostDriver(self.log.info)
    try:
      dpath = STORE % (hd.volroot, cobj._sliver.uuid, "")
      shutil.rmtree(dpath)
    except OSError:
      self.log.info("Could not delete %s" % (dpath))

  def reserveAddress (self, options, user_urn):
    self.log.debug("reserveAddress (%s, %s)" % (self.container._sliver.uuid, self.container.client_id))

    hd = HostDriver(self.log.info)
    dpath = STORE % (hd.volroot, self.container._sliver.uuid, self.container.client_id)
    cpath = "%s/dhcpd.conf" % (dpath)

    hpath = HDATA % (hd.volroot, self.container._sliver.uuid, self.container.client_id)
    hinfo = json.loads(open(hpath, "r").read())
    for entry in options["entries"]:
      hinfo[entry["name"]] = entry

    self._writeConfig(hpath, cpath, hinfo)
    dd = DockerDriver(self.log.info)
    dd.supervisor(self.container, "restart", "dhcpd")
    

  def getLeases (self, options, user_urn):
    self.log.debug("getLeases (%s, %s)" % (self.container._sliver.uuid, self.container.client_id))

    hd = HostDriver(self.log.info)
    leases = IscDhcpLeases("%s/dhcpd.leases" % (STORE % (hd.volroot, self.container._sliver.uuid,
                                                         self.container.client_id)))
    retl = []
    for lease in leases.get_current().values():
      retl.append({"hostname" : lease.hostname, "mac-address" : lease.ethernet,
                   "ip-address" : lease.ip, "binding-state" : lease.binding_state,
                   "start" : lease.start, "end" : lease.end, "options" : lease.options})

    hpath = HDATA % (hd.volroot, self.container._sliver.uuid, self.container.client_id)
    hinfo = json.loads(open(hpath, "r").read())
    for entry in hinfo.values():
      retl.append({"hostname" : entry["name"], "mac-address" : entry["mac"],
                   "ip-address" : entry["addr"], "binding-state" : "unknown",
                   "start" : None, "end" : datetime.datetime(1900, 1, 1), "options" : None})

    return retl

  def setSubnet (self, options, data, user_urn):
    subnet = IPy.IP(data)
    hd = HostDriver(self.log.info)
    try:
      from foam.vts.drivers.docker import DockerDriver
      self.log.debug("setSubnet (%s, %s)" % (self.container._sliver.uuid, self.container.client_id))
      self.attrs["subnet"] = subnet
 
      dpath = STORE % (hd.volroot, self.container._sliver.uuid, self.container.client_id)
      hpath = HDATA % (hd.volroot, self.container._sliver.uuid, self.container.client_id)

      cpath = "%s/dhcpd.conf" % (dpath)
      self._writeConfig(hpath, cpath)
      self.setIPAddress(self.container, "eth1", subnet[1], subnet.prefixlen())

      dd = DockerDriver(self.log.info)
      dd.supervisor(self.container, "restart", "dhcpd")
      return True
    except Exception:
      self.log.exception("Unhandled exception in setSubnet")
      return False

  def setDNS (self, options, user_urn):
    try:
      from foam.vts.drivers.docker import DockerDriver
      self.log.debug("setDNS (%s, %s)" % (self.container._sliver.uuid, self.container.client_id))
      self.attrs["dns-ip"] = options["dns-ip"]

      dpath = STORE % (self.container._sliver.uuid, self.container.client_id)
      cpath = "%s/dhcpd.conf" % (dpath)
      self._writeConfig(cpath)
      
      dd = DockerDriver(self.log.info)
      dd.supervisor(self.container, "restart", "dhcpd")
      return True
    except Exception:
      self.log.exception("Unhandled exception in setDNS")
      return False

  @staticmethod
  def geniSetup (registry):
    SupervisorHost.geniSetup(registry)
    AlpineHost.geniSetup(registry)

    registry.addPOACallback(*(SimpleDHCPHandler.poaSetSubnet))
    registry.addPOACallback(*(SimpleDHCPHandler.poaGetLeases))
    registry.addPOACallback(*(SimpleDHCPHandler.poaReserveAddress))
    registry.addPOACallback(*(SimpleDHCPHandler.poaSetDNS))

SimpleDHCPHandler.poaSetSubnet = _wrapMapPOA("api:uh.dhcp:set-subnet", SimpleDHCPHandler, "setSubnet")
SimpleDHCPHandler.poaGetLeases = _wrapPOA("api:uh.dhcp:get-leases", SimpleDHCPHandler, "getLeases")
SimpleDHCPHandler.poaReserveAddress = _wrapPOA("api:uh.dhcp:reserve-address", SimpleDHCPHandler, "reserveAddress")
SimpleDHCPHandler.poaSetDNS = _wrapPOA("api:uh.dhcp:set-dns", SimpleDHCPHandler, "setDNS")
