# Copyright (c) 2016-2017  University of Houston

from __future__ import absolute_import

import os
import shutil

# Note: To manipulate the bind zonefiles - using the dnspython toolkit http://www.dnspython.org/
# documentation at: http://www.dnspython.org/docs/1.15.0/
import dns.zone
from dns.rdataclass import *
from dns.rdatatype import *
from dns.exception import DNSException
from foam.vts.images import ImageHandler
from foam.vts.hostdriver import HostDriver
from foam.vts.data import DB, UnknownInstanceError
from foam.vts.drivers.docker import HostPathVolume, DockerDriver
from foam.geni.resources import UnknownSliceError, UnsupportedResourceActionError
from foam.core.exception import CoreException
from IPy import IP

from .bases import AlpineSupervisorHost
from .util import RSYSLOGCONF

STORE = "/tmp/dnsroot/%s/%s"

NAMEDCONF = """//%(topDomainName)s
options {
        directory "/data";
        listen-on { %(ipAddress)s; 127.0.0.1;};
        listen-on-v6 { none; };             
        allow-transfer { none; };
        allow-recursion { any; };
        allow-query { any; };
        allow-query-cache { any; };
};                                       
zone "%(topDomainName)s" IN {
        type master;   
        file "db.topzone";
};                                   
zone "." {
        type master;
        file "db.rootzone";
};
zone "localhost" IN{
        type master;
        file "db.localhost";
};
"""
TOPZONEFILE = """$ORIGIN %(topDomainName)s.
@       1D      IN      SOA     %(vmName)s.%(topDomainName)s. hostmaster.%(topDomainName)s. ( 1 3h 15 1w 3h)
                IN      NS      %(vmName)s.%(topDomainName)s.
%(vmName)s              IN      A       %(ipAddress)s
%(delegationRecords)s

"""
ROOTZONEFILE = """$TTL    86400 ;
$ORIGIN .
.               IN      SOA     %(vmName)s.%(topDomainName)s. hostmaster.%(topDomainName)s. ( 1 3h 15 1w 3h )
                IN      NS      %(vmName)s.%(topDomainName)s.
%(topDomainName)s     IN      NS      %(vmName)s.%(topDomainName)s.
%(vmName)s.%(topDomainName)s  IN      A       %(ipAddress)s
"""
LOCALHOSTZONE = """$TTL	86400 ;
$ORIGIN localhost.
@       1D      IN      SOA     %(vmName)s.%(topDomainName)s. hostmaster.%(topDomainName)s. ( 1 3h 15 1w 3h)
@       1D      IN      NS      @ 
        1D      IN      A       127.0.0.1
"""


class DnsRootConfigError(CoreException):
  def __init__(self, client_id, errorDescription):
    super(DnsRootConfigError, self).__init__()
    self.cid = client_id
    self.description = errorDescription

  def __str__(self):
    return "error in configuring uh.dnsroot node [(%s)]: (%s) " % (self.cid, self.description)


class DNSrootHandler(AlpineSupervisorHost):

  def prebuild (self, cobj):
    '''write configs, mount in /data'''
    self.log.debug("Prebuild (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
    os.makedirs(dpath)
    hd = HostDriver(self.log.info)
    self.volumes.append(HostPathVolume(dpath, "/data"))

  def _writeConfigs (self, dpath, cobj):
    '''write the named.conf and required zone files'''
    configpath = "%s/named.conf" % (dpath)
    rootzonepath = "%s/db.rootzone" % (dpath)
    topzonepath = "%s/db.topzone" % (dpath)
    localhostzonepath = "%s/db.localhost" % (dpath)
    topDomainName = self.attrs["top-domain-name"].strip(" .")
    vmName = cobj.hostname
    myport = cobj.ports.values()[0]
    # TODO: Need to handle having no addresses
    try:
      ipAddress = myport.ipv4addrs[0].split('/')[0] #returns a list: take the first.
    except Exception:
      raise DnsRootConfigError(cobj.client_id, 'IPv4 Address of root DNS server port is not set')
    delegationRecords = "";
    sliver = cobj._sliver
    for container in sliver.containers:
      if container.imagename == "uh.dnsdhcp":
        if container._instance_handler.attrs["root-dns"] == cobj.client_id:
          dnsPort = container.getPort(container._instance_handler.attrs["interface-to-root"])
          nsIPAddress = dnsPort.ipv4addrs[0].split('/')[0]  #returns a list: take the first.
          # not catching execption for this above because we set this value in dnsdhcp
          nsHostname = container.hostname
          subdomain = container._instance_handler.attrs["sub-domain-name"].split('.', 1)[0]
          delegationRecords += "\n" + subdomain + " IN NS " + nsHostname + "." +subdomain+ "." + topDomainName + "."
          delegationRecords += "\n" + nsHostname+"."+subdomain + " IN A " + nsIPAddress
    with open(configpath, "w+") as f:
      f.write(NAMEDCONF % {"ipAddress" : str(ipAddress),
                            "topDomainName" : str(topDomainName)})    
    with open(topzonepath, "w+") as f:
      f.write(TOPZONEFILE % {"ipAddress" : str(ipAddress),
                            "topDomainName" : str(topDomainName),
                            "delegationRecords" : str(delegationRecords),
                            "vmName":str(vmName)})
    with open(rootzonepath, "w+") as f:
      f.write(ROOTZONEFILE % {"ipAddress" : str(ipAddress),
                            "topDomainName" : str(topDomainName),
                            "vmName":str(vmName)})
    with open(localhostzonepath, "w+") as f:
      f.write(LOCALHOSTZONE % {"topDomainName" : str(topDomainName),
                            "vmName":str(vmName)})   

  def postlocaltopo (self, cobj):
    pass

  def postnettopo(self, cobj):
    '''write configs, restart the named server'''
    self.log.debug("post net topo (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
    self._writeConfigs(dpath, cobj)
    dd = DockerDriver(self.log.info)
    dd.supervisor(cobj, "start", "named")

  def postrm (self, cobj):
    self.log.debug("postrm (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
    try:
      dpath = STORE % (cobj._sliver.uuid, "")
      shutil.rmtree(dpath)
    except OSError:
      self.log.info("Could not delete %s" % (dpath))

  @staticmethod
  def geniSetup(registry):
    registry.addPOACallback("vts:uh.dnsroot:add-resource-record", poaAddResourceRecord)
    registry.addPOACallback("vts:uh.dnsroot:delete-resource-record", poaDeleteResourceRecord)
    registry.addPOACallback("vts:uh.dnsroot:get-all-records", poaGetAllRecords)

  def addResourceRecord(self, cobj, recordName, recordType, recordValue, recordTTL=900):
    '''uses dns.zone to add a new resource record into db.topzone file'''
    try:
      self.log.debug("addResourceRecord (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
      dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
      topzonepath = "%s/db.topzone" % (dpath)
      domain = self.attrs["top-domain-name"].strip(" .")
      zone = dns.zone.from_file(topzonepath,domain)
      rec = recordName.strip(' .')
      if rec.endswith(domain):
        rec = rec[:len(rec)-len(domain)-1]
      if rec == "":
        rec = "@"
      rdstype = dns.rdatatype.from_text(recordType)
      rdsclass = 1
      rdsset = zone.find_rdataset(rec, rdtype=rdstype, create=True)
      rsdata = dns.rdata.from_text(rdsclass, rdstype, recordValue, origin=None, relativize=True)
      rdsset.add(rsdata, ttl=recordTTL)
      newzone = topzonepath


      for (name, ttl, rdata) in zone.iterate_rdatas(SOA):
        serial = rdata.serial
        new_serial = serial + 1
        rdata.serial = new_serial

      zone.to_file(newzone)
      dd = DockerDriver(self.log.info)
      dd.supervisor(self.container, "restart", "named")
      return True
    except Exception as e:
      self.log.exception("exception in adding Resource record \n" + str(e))
      return False

  def deleteResourceRecord(self, cobj, recordName, recordType):
    '''uses dns.zone to delete an existing resource record(s) from the db.topzonefile'''
    try:
      self.log.debug("deleteResourceRecord (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
      dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
      topzonepath = "%s/db.topzone" % (dpath)
      configpath = "%s/named.conf" % (dpath)
      domain = ''
      with open(configpath, 'r') as f:
        domain = f.readline().replace('\n', '').strip(' ./')
      zone = dns.zone.from_file(topzonepath, domain)
      rec = recordName.strip(' .')
      if rec.endswith(domain):
        rec = rec[:len(rec) - len(domain) - 1]
      if rec == "":
        rec = "@"
      zone.delete_rdataset(rec, rdtype=recordType)
      zoned = topzonepath
      zone.to_file(zoned)
      dd = DockerDriver(self.log.info)
      dd.supervisor(self.container, "restart", "named")
      return True
    except Exception as e:
      self.log.exception("exception in deleting Resource record \n" + str(e))
      return False

  def getAllRecords(self, cobj):
    try:
      self.log.debug("GetAllRecords (%s, %s)" % (cobj._sliver.uuid, cobj.client_id))
      dpath = STORE % (cobj._sliver.uuid, cobj.client_id)
      topzonepath = "%s/db.topzone" % (dpath)
      domain = self.attrs["top-domain-name"].strip(" .")
      zone = dns.zone.from_file(topzonepath, domain)
      retl = []
      for name, node in zone.nodes.items():
        rdatasets = node.rdatasets
        for rdataset in rdatasets:
          value = ""
          for rdata in rdataset:
            value = value + " " + str(rdata)
          retl.append({"name": str(name), "type": to_text(rdataset.rdtype), "TTL": rdataset.ttl, "value": value.strip()})
      return retl
    except Exception as e:
      self.log.debug("exception getting zonefile contents \n " + str(e))
      return []


def poaAddResourceRecord(action, urns, options, user_urn):
  from foam.vts.geni import SliverBundleProxy

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  ct = sliver.getContainer(options["client-id"])
  if not ct:
    raise UnknownInstanceError(options["client-id"], "uh.dnsroot")
  if isinstance(ct._instance_handler, DNSrootHandler):
    retl = {}
    success = ct._instance_handler.addResourceRecord(ct, options["record-name"],options["record-type"],
                                                     options["record-value"],options["record-ttl"])
    retl[options["client-id"]] = success
  else:
    raise UnsupportedResourceActionError("vts:uh.dnsroot:add-resource-record", options["client-id"])
  return retl

def poaDeleteResourceRecord(action, urns, options, user_urn):
  from foam.vts.geni import SliverBundleProxy

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  ct = sliver.getContainer(options["client-id"])
  if not ct:
    raise UnknownInstanceError(options["client-id"], "uh.dnsroot")
  if isinstance(ct._instance_handler, DNSrootHandler):
    retl = {}
    success = ct._instance_handler.deleteResourceRecord(ct, options["record-name"], options["record-type"])
    retl[options["client-id"]] = success
  else:
    raise UnsupportedResourceActionError("vts:uh.dnsroot:delete-resource-record", options["client-id"])
  return retl

def poaGetAllRecords(action, urns, options, user_urn):
  from foam.vts.geni import SliverBundleProxy

  if len(urns) > 1:
    return {}
  try:
    sliver = SliverBundleProxy(DB.getSliversBySlice(urns[0]))
  except KeyError:
    raise UnknownSliceError(urns[0])
  retl = {}
  for clid in options["client-ids"]:
    ct = sliver.getContainer(clid)
    if not ct:
      raise UnknownInstanceError(clid, "uh.dnsdhcp")
    if isinstance(ct._instance_handler, DNSrootHandler):
      zonfilecontents = ct._instance_handler.getAllRecords(ct)
      retl[clid] = zonfilecontents
    else:
      raise UnsupportedResourceActionError("vts:uh.dnsdhcp:get-all-records", clid)
  return retl
